#include <iostream>
#include <vector>
#include <string>
#include "cl_args.hpp"
#include "cl_args_e.hpp"

arg::CL_Args::CL_Args(int argc, char* argv[])
{
    this->count    = argc - 1;
    this->ch_array = argv;
    
    this->setoutput    = false;
    this->fromlist     = false;
    this->frombasename = false;
    this->verbose      = false;
    this->help         = false;
    this->version      = false;
    this->sort         = true;
    
    this->switches.push_back("-b");
    this->switches.push_back("-f");
    this->switches.push_back("-h");
    this->switches.push_back("--help");
    this->switches.push_back("-o");
    this->switches.push_back("-v");
    this->switches.push_back("--version");

    // these strings need to be initialized to an empty string
    // for testing of command line switches
    this->outfile = ""; 
    this->listfile = "";
    this->basename = "";
}

void arg::CL_Args::readCL() throw()
{
    if (this->count > 0) {
	// ch_array is basically "char* argv[]"
	for (int i = 1; i <= this->count; i++){
	    arguments.push_back(ch_array[i]);
	}
	
	try {
	    this->parseArgs();
	    this->argsOK();
	} catch (arg::ArgsError& e) {
	    throw;
	}
    }
}

void arg::CL_Args::parseArgs() throw()
{
    std::vector<std::string>::iterator s;
    
    /* parse for switches that don't need a user input value first */

    // look for and set the "verbose" switch
    s = find(this->arguments.begin(), 
	     this->arguments.end(), "-v");
    (s == this->arguments.end()) ? this->verbose = false : this->verbose = true;

    // look for and set the "version" switch
    s = find(this->arguments.begin(), 
	     this->arguments.end(), "--version");
    (s == this->arguments.end()) ? this->version = false : this->version = true;

    // look for and set the "sort" switch. using "-s" will turn off sorting
    s = find(this->arguments.begin(), 
	     this->arguments.end(), "-s");
    (s == this->arguments.end()) ? this->sort = true : this->sort = false;
    
    // look for and set the "help" switch
    s = find(this->arguments.begin(),
	     this->arguments.end(), "-h");
    (s == this->arguments.end()) ? this->help = false : this->help = true;
    if (this->help == false) {
	s = find(this->arguments.begin(),
		 this->arguments.end(), "--help");
	(s == this->arguments.end()) ? this->help = false : this->help = true;
    }

    /* parse for arguments that users use to set values */

    // look for and set an output value. that is, name the df3 output file
    s = find(this->arguments.begin(),
	     this->arguments.end(), "-o");
    (s == this->arguments.end()) ? this->setoutput = false : this->setoutput = true;
    if (this->setoutput) {
	s++;
	if ( (s == this->arguments.end()) || (this->isSwitch(*s))) { 
	    throw SetOutput();
	} else {
	    this->outfile = *s;
	}
    }

    // look for and set a name that holds a list of files to create a df3 from
    s = find(this->arguments.begin(),
	     this->arguments.end(), "-f");
    (s == this->arguments.end()) ? this->fromlist = false : this->fromlist = true;
    if (this->fromlist) {
	s++;
	if ( (s == this->arguments.end()) || (this->isSwitch(*s))) { 
	    throw FromList();
	} else {
	    this->listfile = *s;
	}
    }

    // look for and set a basename, used to read files in the current directory
    // while you can set this, it's functionality is currently unimplemented
    s = find(this->arguments.begin(),
	     this->arguments.end(), "-b");
    (s == this->arguments.end()) ? this->frombasename = false : this->frombasename = true;
    if (this->frombasename) {
	s++;
	if ( (s == this->arguments.end()) || (this->isSwitch(*s))) { 
	    // why not throw an exception? input was basically: tga2df3 -b
	    // without a value given for "-b"
	    this->basename = "";
	} else {
	    this->basename = *s;
	}
    }
    
    // if there were characters read from the command line 
    // but none were valid options, throw an error
    if ( (this->count > 0) && this->noValidSwitches()){
	throw ArgsError();
    }
}

bool arg::CL_Args::usingOutfile() const
{
    if (this->setOutput() && (this->Outfile() != ""))
	return true;
    return false;
}

bool arg::CL_Args::usingListfile() const
{
    if (this->fromList() && (this->Listfile() != ""))
	return true;
    return false;
}

bool arg::CL_Args::usingBasename() const
{
    if (this->fromBasename() && (this->Basename() != ""))
	return true;
    return false;
}

bool arg::CL_Args::isSwitch(std::string str)
{
    // when parsing arguments that must have a value,
    // check to see if that value is actually another
    // flag instead of a value
    std::vector<std::string>::iterator s;
    
    s = find(this->switches.begin(), 
	     this->switches.end(), str);
    if (s == this->switches.end()) {
	return false;
    }
    return true;
}

bool arg::CL_Args::noValidSwitches()
{
    if ( (!this->setOutput()) &&
	 (!this->fromList()) &&
	 (!this->fromBasename()) &&
	 (!this->Verbose()) &&
	 (!this->Help()) &&
	 (this->Sort()) &&
	 (!this->Version()) ) {
	return true;
    }
    return false;
}

bool arg::CL_Args::argsOK() throw() 
{
    // can only specify one method of getting filenames 
    if (this->usingListfile() && this->usingBasename()){
	throw Input();
    }
    
    // basename is not implemented. throw exception if user tries to use it
    if (this->usingBasename() || this->frombasename){
	throw FromBasename();
    }
    return true;
}

void arg::CL_Args::printAll() const
{
    std::cout << "setoutput:    " << this->setOutput() << "\n"
	      << "fromlist:     " << this->fromList() << "\n"
	      << "frombasename: " << this->fromBasename() << "\n"
	      << "verbose:      " << this->Verbose() << "\n"
	      << "help:         " << this->Help() << "\n"
	      << "sort:         " << this->Sort() << "\n"
	      << "outfile:      " << this->Outfile() << "\n"
	      << "listfile:     " << this->Listfile() << "\n"
	      << "basename:     " << this->Basename() << "\n";
}

void arg::CL_Args::printHelp() const
{
    std::cout << "Usage: tga2df3 [OPTION]\n" 
	      << "Options include:\n"
	/*	      << "  -b                *** NOT YET IMPLEMENTED *** Use a basename to\n"
	      << "                    read TGA files from current directory. \n"
	*/
	      << "  -f                Read TGA filenames from a file, each separated\n"
	      << "                    by a new line. The file must be in your current\n"
	      << "                    working directory or have the full path to tga files\n"
	      << "                    instead of just their names.\n"
	      << "                    Example: tga2df3 -f list.txt\n"
	      << "  -h, --help        Display this help and exit.\n" 
	      << "  -o                Set the output DF3 name. Otherwise filename\n" 
	      << "                    is derived from first TGA name. Example:\n" 
	      << "                    tga2df3 [OPTIONS] -o output.df3\n"
	      << "  -s                Do not sort. Incomming files are sorted by default.\n" 
	      << "  -v                Verbose messages\n"
	      << "  --version         Print version number and exit.\n\n"
	      << "Possibly useful command line examples:\n"
	      << "DOS:\n"
	      << "dir /B /ON your_tga*.tga | tga2df3.exe -o my.df3 -v\n\n"
	      << "Linux:\n"
	      << "ls -1p your_tga*.tga | tga2df3 -o my.df3 -v\n";
}

void arg::CL_Args::printVersion(int major, int minor) const
{
    std::cout << "tga2df3 version " << major << "." << minor << "\n";
}

void arg::CL_Args::debug() const
{
    std::cout << "outfile: " << this->outfile << "\n"
	      << "listfile: " << this->listfile << "\n"
	      << "basename: " << this->basename << "\n";
}
