#include <iostream>
#include <ios>
#include <fstream>
#include "df3.hpp"
#include "tga.hpp"
#include "DataBuffer.hpp"
#include "cl_args.hpp"
#include "tga_e.hpp"

/* if we read a tga file that has properties 
   difference from the previous tga files, this 
   function will return true, meaning the state
   of the DF3 object is inconsistent */
bool df3::DF3::inconsistant(tga::TGA* tga)
{
    return ( ( (this->width != tga->getWidth()) ||
	       (this->height != tga->getHeight()) ||
	       (this->tgatype != tga->getType()) ) &&
	     ( (this->width != -1) &&
	       (this->height != -1) &&
	       (this->tgatype != -1) ) );
}

/* write the binary representation of an integer value
   to df3 file. */
void df3::DF3::writeInt(int value)
{
    unsigned char byte;
    byte = value >> 8;
    this->file << byte;

    byte = (value & 0xFF);
    this->file << byte;
}

bool df3::DF3::DF3isNew()
{
    return ( (this->header  == -1) && 
	     (this->width   == -1) &&
	     (this->height  == -1) &&
	     (this->tgatype == -1) );
}

int df3::DF3::writeDF3Header(tga::TGA* tgafile)
{
    this->width = tgafile->getWidth();
    this->height = tgafile->getHeight();
    
    /* tgatype is not written to DF3, but 
       is set here for completeness */
    this->tgatype = tgafile->getType();	//is it used anywhere?
    
    /* write the DF3 header */
    this->writeInt(this->width);
    this->writeInt(this->height);
    this->writeInt(this->depth);
    
    /* record pixel count for an individual TGA file */
    this->pixels = this->width * this->height;

    /* set header as written */
    this->header = 1;
    return 1;
}

void df3::DF3::tga2df3(std::string file, arg::CL_Args* args) throw()
    //throw (df3::MismatchError, df3::OutputError, tga::TypeError)
{
    try {
	// the constructor may throw an exception. for some reason,
	// this exception isn't unwinding to main()'s catch block
	// so wrap in in a try/catch here and exit if caught
	tga::TGA tgafile(file);
    
	tgafile.readHeader();
	
	if (args->Verbose() && (this->DF3isNew()) ) {
	    std::cout << "DF3 Depth  : " << this->depth << "\n"
		      << "TGA Width  : " << tgafile.getWidth() << "\n"
		      << "TGA Height : " << tgafile.getHeight() << "\n";
	}
	
	if (args->Verbose()) { 
	    std::cout << "Processing file " << file << "\n"; 
	}
	
	/* if current TGA file has a different header than the rest, 
	   we are unable to create a DF3 */
	if (this->inconsistant(&tgafile)) {	
	    throw MismatchError(file, this->width, this->height, this->tgatype, 
				tgafile.getWidth(), tgafile.getHeight(), tgafile.getType());
	}
	
	/* write the DF3 header after we know the width, height, 
	   and number(also refered to as DF3 Depth) of TGA files */
	this->DF3isNew() ? this->writeDF3Header(&tgafile) : 0;
	
	/* based on tga header, seek to image data location */
	tgafile.seekToData();
	
	/* semi-arbitrary buffer size, except it should be a multiple of bpp 
	   so we always read in a complete pixel on the last read */
	int tga_buffer = 32*256*tgafile.getBytesPerPixel();
	
	/* several bytes of tga yeild 1 voxel. so for every 
	   tga_buffer we have df3_buffer worth of data */
	int df3_buffer = tga_buffer / tgafile.getBytesPerPixel(); 
	
	DataBuffer image_data(tga_buffer);
	DataBuffer DF3_frame(df3_buffer);
	
	this->voxelize(DF3_frame, image_data, tgafile);
    } 
    catch (tga::IOError &e) { 
	// for some reason, this exception isn't being unwound properly 
	// to main()'s catch block. so catch it here and exit.
	e.debug();
	exit(1);
    } 
}

void df3::DF3::voxelize(DataBuffer& DF3_frame, DataBuffer& image_data, tga::TGA& tgafile)
{
    // we will read a full buffer of data each time except for the last
    int tga_readcount = this->pixels * tgafile.getBytesPerPixel() / image_data.size();
    int tga_lastread  = this->pixels * tgafile.getBytesPerPixel() % image_data.size();

    int bytes;
    int voxel_count;
    double voxel;
    unsigned char rgb_byte;


    for (int r = 0; r <= tga_readcount; r++){
	// determine if this is the last read that will occur 
	// and set "bytes" accordingly. 
	((r+1) == tga_readcount) ? bytes = tga_lastread : bytes = image_data.size();
	tgafile.readData(image_data, bytes);
	
	voxel_count = 0;
	for (int pos = 0; pos < image_data.size(); pos++){
	    // an individual "voxel" is created by using the greyscaled
	    // value of red, green, and blue bytes 
	    rgb_byte = image_data[pos];
	    voxel = rgb_byte*0.11;
	    pos++;
	    
	    rgb_byte = image_data[pos];
	    voxel += rgb_byte*0.59;
	    pos++;
	    
	    rgb_byte = image_data[pos];
	    voxel += rgb_byte*0.30;
	    
	    // put the value into a buffer. buffer wil be written to disk when finished 
	    DF3_frame[voxel_count++] = static_cast<unsigned char>(voxel);
	}
	// DF3_frame is a buffer of voxels. now that the buffer is 
	// full, write it to disk.
	DF3_frame.writeFile(this->file, bytes/tgafile.getBytesPerPixel());
	
	// check here that write was ok. i feel that it is failing
	// silently in the situation where -o is used to specify
	// a file in a directory that the user doesn't have
	// sufficient permission. this fixes that 
	if (!this->file.good()){
	    throw OutputError(this->name);
	}
    }
}

void df3::DF3::debug() const
{
    std::cout << this->width << " " << this->height << " " << this->depth << "\n"
	      << this->tgatype << " " << this->pixels << " " << this->header << "\n";
}

